#include <ruby.h>
#include "java.h"

static ID id_read, id_readpartial;
static VALUE token_symbol_list[
#define LANGSCAN_JAVA_TOKEN(token) 1 +
  1 + LANGSCAN_JAVA_TOKEN_LIST 0
#undef LANGSCAN_JAVA_TOKEN
];

static void tokenizer_mark(langscan_java_tokenizer_t *tokenizer)
{
  if (tokenizer == NULL)
    return;
  rb_gc_mark((VALUE)langscan_java_tokenizer_get_user_data(tokenizer));
}

static void tokenizer_free(langscan_java_tokenizer_t *tokenizer)
{
  if (tokenizer == NULL)
    return;
  langscan_java_free_tokenizer(tokenizer);
}

static VALUE tokenizer_s_allocate(VALUE klass)
{
  return Data_Wrap_Struct(klass, tokenizer_mark, tokenizer_free, NULL);
}

static size_t user_read_io(void **user_data_p, char *buf, size_t maxlen)
{
  VALUE user_io = (VALUE)*user_data_p;
  VALUE str = rb_funcall(user_io, id_read, 1, INT2NUM(maxlen));
  if (NIL_P(str))
    return 0;
  size_t len = maxlen;
  StringValue(str);
  if (RSTRING(str)->len < len)
    len = RSTRING(str)->len;
  memcpy(buf, RSTRING(str)->ptr, len);
  return len;
}

static size_t user_read_str(void **user_data_p, char *buf, size_t maxlen)
{
  VALUE user_str = (VALUE)*user_data_p;
  StringValue(user_str);
  if (!FL_TEST(user_str, ELTS_SHARED)) {
    user_str = rb_str_new3(rb_str_new4(user_str));
    *user_data_p = (void *)user_str;
  }
  if (RSTRING(user_str)->len < maxlen) {
    maxlen = RSTRING(user_str)->len;
  }
  memcpy(buf, RSTRING(user_str)->ptr, maxlen);
  RSTRING(user_str)->ptr += maxlen;
  RSTRING(user_str)->len -= maxlen;
  return maxlen;
}

static VALUE tokenizer_initialize(VALUE self, VALUE user_data)
{
  VALUE tmp;
  user_read_t user_read;
  langscan_java_tokenizer_t *tokenizer;
  Data_Get_Struct(self, langscan_java_tokenizer_t, tokenizer);
  if (tokenizer) { rb_raise(rb_eArgError, "called twice"); }
  tmp = rb_check_convert_type(user_data, T_FILE, "IO", "to_io");
  if (NIL_P(tmp)) {
    StringValue(user_data);
    user_read = user_read_str;
    user_data = rb_str_new3(rb_str_new4(user_data));
  }
  else {
    user_read = user_read_io;
  }
  DATA_PTR(self) = langscan_java_make_tokenizer(user_read, (void *)user_data);
  return self;
}

static VALUE tokenizer_get_token(VALUE self)
{
  langscan_java_tokenizer_t *tokenizer;
  langscan_java_token_t token;
  Data_Get_Struct(self, langscan_java_tokenizer_t, tokenizer);
  if (tokenizer == NULL) { return Qnil; }
  token = langscan_java_get_token(tokenizer);
  if (token == langscan_java_eof) {
    DATA_PTR(self) = NULL;
    langscan_java_free_tokenizer(tokenizer);
    return Qnil;
  }
  return rb_ary_new3(8,
    token_symbol_list[token],
    rb_str_new(langscan_java_curtoken_text(tokenizer), langscan_java_curtoken_leng(tokenizer)),
    INT2NUM(langscan_java_curtoken_beg_lineno(tokenizer)),
    INT2NUM(langscan_java_curtoken_beg_columnno(tokenizer)),
    INT2NUM(langscan_java_curtoken_beg_byteno(tokenizer)),
    INT2NUM(langscan_java_curtoken_end_lineno(tokenizer)),
    INT2NUM(langscan_java_curtoken_end_columnno(tokenizer)),
    INT2NUM(langscan_java_curtoken_end_byteno(tokenizer)));
}

static VALUE tokenizer_close(VALUE self)
{
  langscan_java_tokenizer_t *tokenizer;
  Data_Get_Struct(self, langscan_java_tokenizer_t, tokenizer);
  if (tokenizer == NULL) { return Qnil; }
  DATA_PTR(self) = NULL;
  langscan_java_free_tokenizer(tokenizer);
  return Qnil;
}

void Init_java()
{
  VALUE LangScan = rb_define_module("LangScan");
  VALUE LangScan_Java = rb_define_module_under(LangScan, "Java");
  VALUE Tokenizer = rb_define_class_under(LangScan_Java, "Tokenizer", rb_cData);
  langscan_java_token_t token_id;

  id_read = rb_intern("read");
  id_readpartial = rb_intern("readpartial");
  token_id = 0;
  token_symbol_list[token_id++] = Qnil;
#define LANGSCAN_JAVA_TOKEN(token) token_symbol_list[token_id++] = ID2SYM(rb_intern(#token));
  LANGSCAN_JAVA_TOKEN_LIST
#undef LANGSCAN_JAVA_TOKEN

  rb_define_alloc_func(Tokenizer, tokenizer_s_allocate);
  rb_define_method(Tokenizer, "initialize", tokenizer_initialize, 1);
  rb_define_method(Tokenizer, "get_token", tokenizer_get_token, 0);
  rb_define_method(Tokenizer, "close", tokenizer_close, 0);
}
